"""Sides - A dice rolling library for games."""

from .constants import d2, d3, d4, d6, d8, d10, d12, d20, d100
from .constants import four_d6, three_d6, two_d6
from .dice import d, Dice, mod, nd
from .find_version import find_version
from .mods import plus, minus, times
from .op import Op
from .rng import AscendingRNG, FixedRNG, high, low, mid, PRNG, RNG, rng
from .roll import max_roll, min_roll, roll


__version__ = find_version()
__author__ = 'Don McCaughey'
__email__ = 'don@donm.cc'



from pathlib import Path


def find_version() -> str:
    file_path = Path(__file__)
    try:
        import importlib.metadata

        package_name = file_path.parent.name
        return importlib.metadata.version(package_name)
    except:
        try:
            import tomllib

            root = file_path.parents[2]
            pyproject_path = root / 'pyproject.toml'
            pyproject_toml = pyproject_path.read_text()
            pyproject = tomllib.loads(pyproject_toml)
            return pyproject['project']['version']
        except:
            return '0.0.0+unknown'


if __name__ == '__main__':
    print(find_version())



from .dice import Dice
from .rng import high, low, RNG


def roll(rng: RNG, *dice_expression: Dice) -> int:
    total = 0
    for dice_term in dice_expression:
        total = dice_term.op(total, sum(dice_term.roll_each_die(rng)))
    return total


def min_roll(*dice_expression: Dice) -> int:
    return roll(low, *dice_expression)


def max_roll(*dice_expression: Dice) -> int:
    return roll(high, *dice_expression)



from __future__ import annotations

import inspect

from collections import OrderedDict
from typing import Any, TypeVar


T = TypeVar('T')


def interned(cls: type[T]) -> type[T]:
    cache: dict[tuple, T] = {}
    original_new = cls.__new__

    def __new__(cls, *args, **kwargs):
        arguments = _get_arguments(cls, *args, **kwargs)

        key = tuple(arguments.values())
        if key in cache:
            return cache[key]

        instance = original_new(cls)
        for name, value in arguments.items():
            object.__setattr__(instance, name, value)

        cache[key] = instance
        return instance

    cls.__new__ = __new__  # type: ignore[method-assign]
    return cls


def _get_arguments(cls: type, *args, **kwargs) -> OrderedDict[str, Any]:
    init_signature = inspect.signature(cls.__init__)  # type: ignore[misc]
    bound_arguments = init_signature.bind(None, *args, **kwargs)
    bound_arguments.apply_defaults()
    bound_arguments.arguments.pop('self')
    return bound_arguments.arguments



from enum import Enum
from functools import total_ordering
from typing import Self, Callable


type BinaryFunction = Callable[[int, int], int]


@total_ordering
class Op(Enum):
    PLUS = '+', int.__add__
    MINUS = '-', int.__sub__
    TIMES = 'x', int.__mul__

    def __init__(self, symbol: str, f: BinaryFunction):
        self.symbol = symbol
        self.f = f

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}.{self.name}>'

    def __str__(self) -> str:
        return self.symbol

    def __lt__(self, other: Self) -> bool:
        assert isinstance(other, Op)
        return self.symbol < other.symbol

    def __call__(self, a: int, b: int) -> int:
        return self.f(a, b)
